# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import typing

import numpy as np

from dimod.binary_quadratic_model import BinaryQuadraticModel
from dimod.constrained import ConstrainedQuadraticModel

__all__ = ['random_knapsack']


def random_knapsack(num_items: int,
                    seed: typing.Optional[int] = None,
                    value_range: typing.Tuple[int, int] = (10, 30),
                    weight_range: typing.Tuple[int, int] = (10, 30),
                    tightness_ratio: float = 0.5,
                    ) -> ConstrainedQuadraticModel:
    """Generates a constrained quadratic model encoding a knapsack problem.

    Given the number of items, generates a random knapsack problem, formulated as
    a :class:`~dimod.ConstrainedQuadraticModel`. The capacity of bins is set
    to be ``tightness_ratio`` times the sum of the weights.

    Args:
        num_items: Number of items to choose from.

        seed: Seed for NumPy random number generator.

        value_range: Range of the randomly generated values for each item.

        weight_range: Range of the randomly generated weights for each item.

        tightness_ratio: Ratio of capacity over sum of weights.

    Returns:
        The quadratic model encoding the knapsack problem. Variables are
        denoted as ``x_{i}`` where ``x_{i} == 1`` means that item ``i`` is
        placed in the knapsack.

    """

    rng = np.random.RandomState(seed)

    value = {i: rng.randint(*value_range) for i in range(num_items)}
    weight = {i: rng.randint(*weight_range) for i in range(num_items)}
    capacity = int(sum(weight.values()) * tightness_ratio)

    model = ConstrainedQuadraticModel()
    obj = BinaryQuadraticModel(vartype='BINARY')
    x = {i: obj.add_variable(f'x_{i}') for i in range(num_items)}

    for i in range(num_items):
        obj.set_linear(x[i], -value[i])

    model.set_objective(obj)
    constraint = [(x[i], weight[i]) for i in range(num_items)] + [(-capacity, )]
    model.add_constraint(constraint, sense="<=", label='capacity')

    return model



# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Type hints for common dimod inputs.
"""
import typing

import numpy as np

from dimod.vartypes import VartypeLike

try:
    from numpy.typing import ArrayLike, DTypeLike
except ImportError:
    # support numpy < 1.20
    ArrayLike = typing.Sequence
    DTypeLike = typing.Any

try:
    from numpy.typing import NDArray
except ImportError:
    # support numpy < 1.21
    NDArray = typing.Sequence


__all__ = ['Bias',
           'GraphLike',
           'Polynomial',
           'SampleLike',
           'SamplesLike',
           'Variable',
           'VartypeLike',
           ]

# use float for python types, https://www.python.org/dev/peps/pep-0484/#the-numeric-tower
# exclude np.complexfloating from numpy types
Bias = typing.Union[float, np.floating, np.integer]
"""A :obj:`~typing.Union` representing objects that can be used as biases.

This includes:

* Python's :class:`int` and :class:`float`.
* NumPy's :class:`~numpy.floating` and :class:`~numpy.integer`.

"""

# it would be nice to be able to exclude None from Variable. We can define
# a special abstract base class that excludes None, but making it behave like
# a typing class is a pain accross our supported Python versions. In the future
# we should handle it

Variable = typing.Hashable
"""Objects that can be used as variable labels.

.. note::

    In `dimod` variables can be labelled using any hashable object except for
    :obj:`None`. However, for simplicity we alias :class:`~typing.Hashable` which
    does permit :obj:`None`.

"""


GraphLike = typing.Union[
    int,  # number of nodes
    typing.Tuple[typing.Collection[Variable], typing.Collection[typing.Tuple[Variable, Variable]]],
    typing.Collection[typing.Tuple[Variable, Variable]],  # edges
    ]
"""Objects that can be interpreted as a graph.

This includes:

* An :class:`int`, interpreted as a complete graph with nodes labelled ``range(n)``.
* A list of edges
* A 2-tuple containing a list of nodes and a list of edges
* A :class:`networkx.Graph`.

"""

try:
    import networkx as nx
except ImportError:
    pass
else:
    GraphLike = typing.Union[GraphLike, nx.Graph]

Polynomial = typing.Mapping[typing.Sequence[Variable], Bias]
"""A polynomial represented by a mapping."""


class QuadraticVectors(typing.NamedTuple):
    row_indices: NDArray[np.integer]
    col_indices: NDArray[np.integer]
    biases: NDArray[np.floating]


class BQMVectors(typing.NamedTuple):
    linear_biases: NDArray[np.floating]
    quadratic: QuadraticVectors
    offset: Bias


class LabelledBQMVectors(typing.NamedTuple):
    linear_biases: NDArray[np.floating]
    quadratic: QuadraticVectors
    offset: Bias
    labels: typing.Sequence[Variable]


class DQMVectors(typing.NamedTuple):
    case_starts: NDArray[np.integer]
    linear_biases: NDArray[np.floating]
    quadratic: QuadraticVectors
    labels: typing.Sequence[Variable]
    offset: Bias


SampleLike = typing.Union[
    typing.Sequence[float],
    typing.Mapping[Variable, Bias],
    ArrayLike,  # this is overgenerous, but we cannot easily specify it better
    ]
"""Objects that can be interpreted as a single sample.

This includes:

* A one-dimensional NumPy array_like_.
* A 2-:class:`tuple` containing a one-dimensional NumPy array_like_ and a
  list of variable labels.
* A :class:`dict` where the keys are variable labels and the values are the
  assignments.

NumPy array_like_ is a very flexible definition.

.. _array_like: https://numpy.org/devdocs/glossary.html#term-array_like

"""


SamplesLike = typing.Union[
    SampleLike,
    typing.Sequence[typing.Sequence[float]],  # 2d array
    typing.Tuple[typing.Sequence[float], typing.List[Variable]],
    typing.Tuple[typing.Sequence[typing.Sequence[float]], typing.List[Variable]],
    typing.Sequence[SampleLike],
    typing.Iterator[SampleLike],
    ]
"""Objects that can be interpreted as a collection of samples.

This includes:

* Any :obj:`SampleLike`.
* A two-dimensional NumPy array_like_.
* A 2-:class:`tuple` containing a two-dimensional array_like_ and a
  list of variable labels.
* A :class:`list` of :class:`dict` where each dict has the same keys.

.. _array_like: https://numpy.org/devdocs/glossary.html#term-array_like

"""



# Copyright 2022 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from __future__ import annotations

import collections.abc
import functools
import io
import os
import shutil
import tempfile
import typing

import dimod  # for typing

from dimod.cylp import cyread_lp_file


__all__ = ['loads', 'load']


def load(file_like: typing.Union[str, bytes, io.IOBase]) -> dimod.ConstrainedQuadraticModel:
    """Construct a constrained quadratic model from a LP file.

    LP files are a common format for encoding optimization models. See
    documentation from Gurobi_ and CPLEX_.

    Args:
        file_like: Either a :class:`str` or :class:`bytes` object representing
            a path, or a file-like_ object.

    Returns:

        An example of reading from a LP file.

        .. code-block:: python

            with open('example.lp', 'rb') as f:
                cqm = dimod.lp.load(f)

        An example of loading from a path

        .. code-block:: python

            cqm = dimod.lp.load('example.lp')

    See also:

        :func:`~dimod.serialization.lp.loads`

    .. versionadded:: 0.11.0

    .. _Gurobi: https://www.gurobi.com/documentation/9.5/refman/lp_format.html

    .. _CPLEX: https://www.ibm.com/docs/en/icos/12.8.0.0?topic=cplex-lp-file-format-algebraic-representation

    .. _file-like: https://docs.python.org/3/glossary.html#term-file-object

    """

    if isinstance(file_like, (str, bytes)):
        return cyread_lp_file(file_like)

    # ok, we got a file-like

    if not file_like.readable():
        raise ValueError("file_like must be readable")

    if file_like.seekable() and file_like.tell():
        # this file is current pointing somewhere other than the beginning,
        # so we make a copy so our reader can start at the beginning
        filename = ''
    else:
        try:
            filename = file_like.name
        except AttributeError:
            filename = ''

    if os.path.isfile(filename):
        return cyread_lp_file(filename)        

    # copy it into somewhere that our reader can get it
    with tempfile.NamedTemporaryFile('wb', delete=False) as tf:
        shutil.copyfileobj(file_like, tf)

    try:
        return cyread_lp_file(tf.name)
    finally:
        # remove the file so we're not accumulating memory/disk space
        os.unlink(tf.name)


def loads(obj: typing.Union[str, typing.ByteString]) -> dimod.ConstrainedQuadraticModel:
    """Construct a constrained quadratic model from a string formatted as a LP file.

    LP files are a common format for encoding optimization models. See
    documentation from Gurobi_ and CPLEX_.

    Args:
        obj: A :class:`str` or :class:`bytes` formatted like an LP file.

    Returns:

        An example of reading a string formatted as an LP file.

        .. testcode::

            lp = '''
            Minimize
                x0 - 2 x1
            Subject To
                x0 + x1 = 1
            Binary
                x0 x1
            End
            '''

            cqm = dimod.lp.loads(lp)


    See also:

        :func:`~dimod.serialization.lp.load`

    .. versionadded:: 0.11.0

    .. _Gurobi: https://www.gurobi.com/documentation/9.5/refman/lp_format.html

    .. _CPLEX: https://www.ibm.com/docs/en/icos/12.8.0.0?topic=cplex-lp-file-format-algebraic-representation

    .. _file-like: https://docs.python.org/3/glossary.html#term-file-object

    """

    if isinstance(obj, str):
        obj = obj.encode()

    return load(io.BytesIO(obj))



# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Unified warnings/exceptions for stuff moved to dwave-preprocessing."""

__all__ = ['DeprecatedToPreprocessing']


class DeprecatedToPreprocessing:
    """Removed.

    .. deprecated:: 0.10.0

        The following composites were migrated to
        `dwave-preprocessing <https://github.com/dwavesystems/dwave-preprocessing>`_

        * ``ClipComposite``
        * ``ConnectedComponentsComposite``
        * ``FixedVariableComposite``
        * ``RoofDualityComposite``
        * ``ScaleComposite``
        * ``SpinReversalTransformComposite``

        You must install ``dwave-preprocessing`` in order to use them.

    """
    def __init__(self, *args, **kwargs):
        raise TypeError(
                f"{type(self).__name__!r} was moved to dwave-preprocessing in dimod 0.10.0. "
                "You must install dwave-preprocessing in order to use it. "
                "You can do so with 'pip install dwave-preprocessing'. "
                "This stub class will be removed in dimod 0.13.0.",
            )



# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from numbers import Number

from cpython.long cimport PyLong_Check
from cpython.dict cimport PyDict_Size, PyDict_Contains
from cpython.ref cimport PyObject

from dimod.utilities import iter_safe_relabels

cdef extern from "Python.h":
    # not yet available as of cython 0.29.22
    PyObject* PyDict_GetItemWithError(object p, object key) except? NULL


cdef class cyVariables:
    def __init__(self, object iterable=None):
        self._index_to_label = dict()
        self._label_to_index = dict()
        self._stop = 0

        if iterable is not None:
            for v in iterable:
                self._append(v, permissive=True)

    def __contains__(self, v):
        return bool(self.count(v))

    def __copy__(self):
        return self.copy()
    
    # we can almost make __deepcopy__ just call __copy__ because
    # most potential variable types (str, int, tuple) are atomic. However
    # atomic is a strict subset of hashable (e.g. frozenset), so we fall back
    # on the default __deepcopy__

    def __getitem__(self, idx):
        try:
            return self.at(idx)
        except TypeError:
            pass

        if not isinstance(idx, slice):
            raise TypeError(f"indices must be integers or slices, not {type(idx)}")

        cdef Py_ssize_t start = 0 if idx.start is None else idx.start
        cdef Py_ssize_t stop = self.size() if idx.stop is None else idx.stop
        cdef Py_ssize_t step = 1 if idx.step is None else idx.step

        cdef Py_ssize_t i
        cdef cyVariables new = type(self)()
        for i in range(start, stop, step):
            new._append(self.at(i), permissive=False)
        return new

    def __iter__(self):
        cdef Py_ssize_t i

        if self._is_range():
            yield from range(self._stop)
        else:
            for i in range(self._stop):
                yield self.at(i)

    def __len__(self):
        return self.size()

    cpdef object _append(self, object v=None, bint permissive=False):
        """Append a new variable.

        Args:
            v (hashable, optional):
                Add a new variable. If `None`, a new label will be generated.
                The generated label will be the index of the new variable if
                that index is available, otherwise it will be the lowest
                available non-negative integer.

            permissive (bool, optional, default=False):
                If `False`, appending a variable that already exists will raise
                a `ValueError`. If `True`, appending a variable that already
                exists will not change the container.

        Returns:
            hashable: The label of the appended variable.

        Raises:
            ValueError: If the variable is present and `permissive` is
            False.

        This method is semi-public. it is intended to be used by
        classes that have :class:`.Variables` as an attribute, not by the
        the user.
        """
        if v is None:
            v = self._stop

            if not self._is_range() and self.count(v):
                v = 0
                while self.count(v):
                    v += 1

        elif self.count(v):
            if permissive:
                return v
            else:
                raise ValueError('{!r} is already a variable'.format(v))

        idx = self._stop

        if idx != v:
            self._label_to_index[v] = idx
            self._index_to_label[idx] = v

        self._stop += 1
        return v

    cpdef bint _is_range(self):
        """Return whether the variables are currently labelled [0, n)."""
        return not PyDict_Size(self._label_to_index)

    cpdef object _extend(self, object iterable, bint permissive=False):
        """Add new variables.

        Args:
            iterable (iterable[hashable], optional):
                An iterable of hashable objects.

            permissive (bool, optional, default=False):
                If `False`, appending a variable that already exists will raise
                a `ValueError`. If `True`, appending a variable that already
                exists will not change the container.

        Raises:
            ValueError: If a variable is present and `permissive` is
            False.

        This method is semi-public. it is intended to be used by
        classes that have :class:`.Variables` as an attribute, not by the
        the user.
        """
        # todo: performance improvements for range etc
        for v in iterable:
            self._append(v, permissive=permissive)

    cpdef object _pop(self):
        """Remove the last variable.

        This method is semi-public. it is intended to be used by
        classes that have :class:`.Variables` as an attribute, not by the
        the user.
        """
        if not self:
            raise IndexError("Cannot pop when Variables is empty")

        self._stop = idx = self._stop - 1

        label = self._index_to_label.pop(idx, idx)
        self._label_to_index.pop(label, None)
        return label

    def _relabel(self, mapping):
        """Relabel the variables in-place.

        Args:
            mapping (dict):
                Mapping from current variable labels to new, as a dict. If
                an incomplete mapping is specified, unmapped variables keep
                their current labels.

        This method is semi-public. it is intended to be used by
        classes that have :class:`.Variables` as an attribute, not by the
        the user.
        """
        for submap in iter_safe_relabels(mapping, self):
            for old, new in submap.items():
                if old == new:
                    continue

                idx = self._label_to_index.pop(old, old)

                if new != idx:
                    self._label_to_index[new] = idx
                    self._index_to_label[idx] = new  # overwrites old idx
                else:
                    self._index_to_label.pop(idx, None)

    def _relabel_as_integers(self):
        """Relabel the variables as integers in-place.

        Returns:
            dict: A mapping that will restore the original labels.

        Examples:

            >>> variables = dimod.variables.Variables(['a', 'b', 'c', 'd'])
            >>> print(variables)
            Variables(['a', 'b', 'c', 'd'])
            >>> mapping = variables._relabel_as_integers()
            >>> print(variables)
            Variables([0, 1, 2, 3])
            >>> variables._relabel(mapping)  # restore the original labels
            >>> print(variables)
            Variables(['a', 'b', 'c', 'd'])

        This method is semi-public. it is intended to be used by
        classes that have :class:`.Variables` as an attribute, not by the
        the user.
        """
        mapping = self._index_to_label.copy()
        self._index_to_label.clear()
        self._label_to_index.clear()
        return mapping

    cdef object at(self, Py_ssize_t idx):
        """Get variable `idx`.

        This method is useful for accessing from cython since __getitem__ goes
        through python.
        """
        if idx < 0:
            idx = self._stop + idx

        if not 0 <= idx < self._stop:
            raise IndexError('index out of range')

        cdef object v
        cdef object pyidx = idx
        cdef PyObject* obj
        if self._is_range():
            v = pyidx
        else:
            # faster than self._index_to_label.get
            obj = PyDict_GetItemWithError(self._index_to_label, pyidx)
            if obj == NULL:
                v = pyidx
            else:
                v = <object>obj  # correctly handles the ref count

        return v

    cpdef cyVariables copy(self):
        """Return a copy of the Variables object."""
        cdef cyVariables new = self.__new__(type(self))
        new._index_to_label = dict(self._index_to_label)
        new._label_to_index = dict(self._label_to_index)
        new._stop = self._stop
        return new

    cdef Py_ssize_t _count_int(self, object v) except -1:
        # only works when v is an int
        cdef Py_ssize_t vi = v

        if self._is_range():
            return 0 <= vi < self._stop

        # need to make sure that we're not using the integer elsewhere
        return (0 <= vi < self._stop
                and not PyDict_Contains(self._index_to_label, v)
                or PyDict_Contains(self._label_to_index, v))

    cpdef Py_ssize_t count(self, object v) except -1:
        """Return the number of times `v` appears in the variables.

        Because the variables are always unique, this will always return 1 or 0.
        """
        if PyLong_Check(v):
            return self._count_int(v)

        # handle other numeric types
        if isinstance(v, Number):
            v_int = int(v)  # assume this is safe because it's a number
            if v_int == v:
                return self._count_int(v_int)  # it's an integer afterall!

        try:
            return v in self._label_to_index
        except TypeError:
            # unhashable
            return False

    cpdef Py_ssize_t index(self, object v, bint permissive=False) except -1:
        """Return the index of `v`.

        Args:
            v (hashable):
                A variable.

            permissive (bool, optional, default=False):
                If True, the variable will be inserted, guaranteeing an index
                can be returned.

        Returns:
            int: The index of the given variable.

        Raises:
            ValueError: If the variable is not present and `permissive` is
            False.

        """
        if permissive:
            self._append(v, permissive=True)
        if not self.count(v):
            raise ValueError('unknown variable {!r}'.format(v))

        if self._is_range():
            return v if PyLong_Check(v) else int(v)

        # faster than self._label_to_index.get
        cdef PyObject* obj = PyDict_GetItemWithError(self._label_to_index, v)
        if obj == NULL:
            pyobj = v
        else:
            pyobj = <object>obj  # correctly updates ref count

        return pyobj if PyLong_Check(pyobj) else int(pyobj)

    cdef Py_ssize_t size(self):
        """The number of variables.

        This method is useful for accessing from cython since __len__ goes
        through python.
        """
        return self._stop



# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import copy
import functools

from collections.abc import Collection, Iterator, Callable, Sequence
from operator import add
from typing import Tuple, Iterator, Optional, Mapping, Any

import numpy as np

try:
    from numpy.typing import ArrayLike, DTypeLike
except ImportError:
    DTypeLike = Any

from dimod.binary.pybqm import pyBQM
from dimod.sampleset import as_samples
from dimod.typing import Bias, Variable
from dimod.vartypes import BINARY, SPIN, Vartype, as_vartype, VartypeLike

__all__ = ['VartypeView']


def view_method(f):
    @functools.wraps(f)
    def wrapper(obj, *args, **kwargs):
        # explicitly check that we're getting the expected vartype combinations
        if obj._vartype == obj.data.vartype():
            # if the vartype matches, just use the underlying data method
            return getattr(obj.data, f.__name__)(*args, **kwargs)

        elif obj.data.vartype() is SPIN and obj._vartype is BINARY:
            return f(obj, *args, **kwargs)

        elif obj.data.vartype() is BINARY and obj._vartype is SPIN:
            return f(obj, *args, **kwargs)

        else:
            raise RuntimeError("unexpected vartype combination")

    return wrapper


class VartypeView:
    def __init__(self, data, vartype: Vartype):
        self.data = data
        self._vartype = vartype

    def __copy__(self):
        # since we'd need to copy the underlying data anyway, let's just return
        # that instead. It doesn't really make sense to maintain the view
        # of a detached copy.
        new = copy.copy(self.data)
        new.change_vartype(self._vartype)
        return new

    @property
    def dtype(self) -> np.dtype:
        return self.data.dtype

    @property
    def variables(self) -> Sequence:
        return self.data.variables

    @property
    def offset(self) -> Bias:
        if self._vartype == self.data.vartype():
            return self.data.offset
        elif self._vartype is BINARY and self.data.vartype() is SPIN:
            # binary <- spin
            return (self.data.offset
                    - self.data.reduce_linear(add, 0)
                    + self.data.reduce_quadratic(add, 0))
        elif self._vartype is SPIN and self.data.vartype() is BINARY:
            # spin <- binary
            return (self.data.offset
                    + self.data.reduce_linear(add, 0) / 2
                    + self.data.reduce_quadratic(add, 0) / 4)
        else:
            raise RuntimeError("unexpected vartype combination")

    @offset.setter
    def offset(self, bias: Bias):
        if self._vartype == self.data.vartype:
            self.data.offset = bias
        elif self._vartype is BINARY and self.data.vartype() is SPIN:
            # binary <- spin
            self.data.offset += bias - self.offset  # use the difference
        elif self._vartype is SPIN and self.data.vartype() is BINARY:
            # spin <- binary
            self.data.offset += bias - self.offset  # use the difference
        else:
            raise RuntimeError("unexpected vartype combination")

    @view_method
    def add_linear(self, v: Variable, bias: Bias):
        if self._vartype is BINARY:  # binary -> spin
            self.data.add_linear(v, bias / 2)
            self.data.offset += bias / 2
        else:  # spin -> binary
            self.data.add_linear(v, 2*bias)
            self.data.offset -= bias

    def add_linear_equality_constraint(self, *args, **kwargs):
        raise NotImplementedError  # defer to caller

    @view_method
    def add_quadratic(self, u: Variable, v: Variable, bias: Bias):
        if self._vartype is BINARY:  # binary -> spin
            self.data.add_quadratic(u, v, bias / 4)
            self.data.add_linear(u, bias / 4)
            self.data.add_linear(v, bias / 4)
            self.data.offset += bias / 4
        else:  # spin -> binary
            self.data.add_quadratic(u, v, 4*bias)
            self.data.add_linear(u, -2*bias)
            self.data.add_linear(v, -2*bias)
            self.data.offset += bias

    def add_variable(self, v: Optional[Variable] = None,
                     bias: Bias = 0) -> Variable:
        v = self.data.add_variable(v)
        self.add_linear(v, bias)
        return v

    def change_vartype(self, vartype: VartypeLike):
        self._vartype = as_vartype(vartype)

    def degree(self, v: Variable):
        return self.data.degree(v)

    @view_method
    def energies(self, samples_like, dtype: DTypeLike = None):
        samples, labels = as_samples(samples_like, copy=True)

        if self._vartype is BINARY:  # binary -> spin
            samples *= 2
            samples -= 1
        else:  # spin -> binary
            samples += 1
            samples //= 2

        return self.data.energies((samples, labels), dtype=dtype)

    @view_method
    def get_linear(self, v: Variable) -> Bias:
        if self._vartype is BINARY:  # binary <- spin
            return (2 * self.data.get_linear(v)
                    - 2 * self.data.reduce_neighborhood(v, add, 0))
        else:  # spin <- binary
            return (self.data.get_linear(v) / 2
                    + self.data.reduce_neighborhood(v, add, 0) / 4)

    @view_method
    def get_quadratic(self, u: Variable, v: Variable,
                      default: Optional[Bias] = None) -> Bias:
        if u == v:
            raise ValueError(f"{u!r} cannot have an interaction with itself")

        try:
            if self._vartype is BINARY:  # binary <- spin
                return 4 * self.data.get_quadratic(u, v)
            else:  # spin <- binary
                return self.data.get_quadratic(u, v) / 4
        except ValueError as err:
            if default is None:
                raise ValueError(
                    f"{u!r} and {v!r} have no interaction") from None
            return default

    def is_linear(self) -> bool:
        return self.data.is_linear()

    @view_method
    def iter_neighborhood(self, v: Variable) -> Iterator[Tuple[Variable, Bias]]:
        if self._vartype is BINARY:  # binary <- spin
            for u, bias in self.data.iter_neighborhood(v):
                yield u, 4 * bias
        else:  # spin <- binary
            for u, bias in self.data.iter_neighborhood(v):
                yield u, bias / 4

    @view_method
    def iter_quadratic(self) -> Iterator[Tuple[Variable, Variable, Bias]]:
        if self._vartype is BINARY:  # binary <- spin
            for u, v, bias in self.data.iter_quadratic():
                yield u, v, 4 * bias
        else:  # spin <- binary
            for u, v, bias in self.data.iter_quadratic():
                yield u, v, bias / 4

    def nbytes(self, capacity: bool = False) -> int:
        return self.data.nbytes(capacity=capacity)

    def num_interactions(self):
        return self.data.num_interactions()

    def num_variables(self):
        return self.data.num_variables()

    def reduce_linear(self, function: Callable,
                      initializer: Optional[Bias] = None) -> Bias:
        gen = (self.get_linear(v) for v in self.variables)
        if initializer is None:
            return functools.reduce(function, gen)
        else:
            return functools.reduce(function, gen, initializer)

    def reduce_neighborhood(self, v: Variable, function: Callable,
                            initializer: Optional[Bias] = None) -> Bias:
        gen = (b for _, b in self.iter_neighborhood(v))
        if initializer is None:
            return functools.reduce(function, gen)
        else:
            return functools.reduce(function, gen, initializer)

    def reduce_quadratic(self, function: Callable,
                         initializer: Optional[Bias] = None) -> Bias:
        gen = (b for _, _, b in self.iter_quadratic())
        if initializer is None:
            return functools.reduce(function, gen)
        else:
            return functools.reduce(function, gen, initializer)

    @view_method
    def remove_interaction(self, u: Variable, v: Variable):
        self.get_quadratic(u, v)  # raise an error if it doesn't exist
        self.set_quadratic(u, v, 0)  # zero it out in the appropriate vartype
        self.data.remove_interaction(u, v)

    @view_method
    def remove_variable(self, v: Optional[Variable] = None) -> Variable:
        if v is None:
            try:
                v = self.variables[-1]
            except IndexError:
                raise ValueError("cannot pop from an empty model")
        # set everything associated with `v` to 0
        for u, _ in self.iter_neighborhood(v):
            self.set_quadratic(u, v, 0)
        self.set_linear(v, 0)
        return self.data.remove_variable(v)

    def relabel_variables(self, mapping: Mapping[Variable, Variable]):
        self.data.relabel_variables(mapping)

    def relabel_variables_as_integers(self) -> Mapping[int, Variable]:
        return self.data.relabel_variables_as_integers()

    @view_method
    def set_linear(self, v: Variable, bias: Bias):
        self.add_linear(v, 0)  # make sure it exists
        self.add_linear(v, bias - self.get_linear(v))  # just add the delta

    def set_quadratic(self, u: Variable, v: Variable, bias: Bias):
        self.add_variable(u)
        self.add_variable(v)
        # just add the delta
        self.add_quadratic(u, v, 0)  # make sure it exists
        self.add_quadratic(u, v, bias - self.get_quadratic(u, v))

    def to_numpy_vectors(self, *args, **kwargs):
        raise NotImplementedError  # defer to the caller

    def update(self, *args, **kwargs):
        raise NotImplementedError  # defer to the caller

    def vartype(self, v: Optional[Variable] = None) -> Vartype:
        return self._vartype



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import typing


def fix_variables(*args, **kwargs) -> typing.NoReturn:
    """Removed

    .. deprecated:: 0.10.0

        This function was migrated to
        `dwave-preprocessing <https://github.com/dwavesystems/dwave-preprocessing>`_

        Use ``dwave.preprocessing.roof_duality`` instead.

    """
    raise TypeError(
        "fix_variables() was moved to dwave-preprocessing in dimod 0.10.0. "
        "Install dwave-preprocessing and use dwave.preprocessing.roof_duality instead. "
        "You can do so with 'pip install dwave-preprocessing'. "
        "This stub function will be removed in dimod 0.13.0.",
        )



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import collections.abc as abc
import inspect
import itertools
import warnings

from functools import wraps
from numbers import Integral

from dimod.exceptions import BinaryQuadraticModelStructureError, WriteableError
from dimod.utilities import new_label
from dimod.vartypes import as_vartype

__all__ = ['nonblocking_sample_method',
           'bqm_index_labels',
           'bqm_index_labelled_input'
           'bqm_structured',
           'vartype_argument',
           'graph_argument',
           'forwarding_method',
           ]


def nonblocking_sample_method(f):
    """Decorator to create non-blocking sample methods.

    Some samplers work asynchronously, and it is useful for composites to
    handle that case. This decorator can be used to easily construct a
    non-blocking :class:`.Sampler` or :class:`.Composite`.

    The function being decorated must return an iterator when called. This
    iterator must yield exactly two values.

    The first value can be any object, but if the object has a `done()`
    method, that method will determine the value of :meth:`.SampleSet.done()`.

    The second value must be a :class:`.SampleSet`, which will provide the
    samples to the user.

    The generator is executed until the first yield. The generator is then
    resumed when the returned sample set is resolved.

    >>> from dimod.decorators import nonblocking_sample_method
    ...
    >>> class Sampler:
    ...     @nonblocking_sample_method
    ...     def sample(self, bqm):
    ...         print("First part!")
    ...         yield
    ...         print("Second part!")
    ...         sample = {v: 1 for v in bqm.variables}
    ...         yield dimod.SampleSet.from_samples_bqm(sample, bqm)
    ...
    >>> bqm = dimod.BinaryQuadraticModel.from_ising({'a': -1}, {('a', 'b'): 1})
    >>> ss = Sampler().sample(bqm)
    First part!
    >>> ss.resolve()
    Second part!
    >>> print(ss)
       a  b energy num_oc.
    0 +1 +1    0.0       1
    ['SPIN', 1 rows, 1 samples, 2 variables]

    """
    from dimod.sampleset import SampleSet  # avoid circular import

    @wraps(f)
    def _sample(*args, **kwargs):
        iterator = f(*args, **kwargs)

        # resolve blocking part now, and make hook for the non-blocking part
        return SampleSet.from_future(next(iterator), lambda _: next(iterator))
    return _sample


def bqm_index_labels(f):
    """Decorator to convert a BQM to index-labels and relabel the sample set
    output.

    Designed to be applied to :meth:`.Sampler.sample`. Expects the wrapped
    function or method to accept a :obj:`.BinaryQuadraticModel` as the second
    input and to return a :obj:`.SampleSet`.

    """
    @wraps(f)
    def _index_label(sampler, bqm, **kwargs):
        bqm, mapping = bqm.relabel_variables_as_integers(inplace=False)
        return f(sampler, bqm, **kwargs).relabel_variables(mapping, inplace=False)
    return _index_label


def bqm_structured(f):
    """Decorator to raise an error if the given BQM does not match the sampler's
    structure.

    Designed to be applied to :meth:`.Sampler.sample`. Expects the wrapped
    function or method to accept a :obj:`.BinaryQuadraticModel` as the second
    input and for the :class:`.Sampler` to also be :class:`.Structured`.
    """
    @wraps(f)
    def structured_sample(sampler, bqm, **kwargs):
        adjacency = sampler.adjacency

        if not adjacency.keys() >= bqm.variables:
            raise BinaryQuadraticModelStructureError(
                f"given bqm has at least one variable, {(list(bqm.variables) - adjacency.keys()).pop()!r}, "
                "not supported by the structured solver")
        if any(u not in adjacency[v] for u, v, _ in bqm.iter_quadratic()):
            u, v = next((u, v) for u, v, _ in bqm.iter_quadratic() if u not in adjacency[v])
            raise BinaryQuadraticModelStructureError(
                f"given bqm contains an interaction, {(u, v)!r}, "
                "not supported by the structured solver")
        return f(sampler, bqm, **kwargs)
    return structured_sample


def vartype_argument(*arg_names):
    """Ensures the wrapped function receives valid vartype argument(s).

    One or more argument names can be specified as a list of string arguments.

    Args:
        *arg_names (list[str], argument names, optional, default='vartype'):
            Names of the constrained arguments in decorated function.

    Returns:
        Function decorator.

    Examples:

        >>> from dimod.decorators import vartype_argument

        >>> @vartype_argument()
        ... def f(x, vartype):
        ...     print(vartype)
        ...
        >>> f(1, 'SPIN')
        Vartype.SPIN
        >>> f(1, vartype='SPIN')
        Vartype.SPIN

        >>> @vartype_argument('y')
        ... def f(x, y):
        ...     print(y)
        ...
        >>> f(1, 'SPIN')
        Vartype.SPIN
        >>> f(1, y='SPIN')
        Vartype.SPIN

        >>> @vartype_argument('z')
        ... def f(x, **kwargs):
        ...     print(kwargs['z'])
        ...
        >>> f(1, z='SPIN')
        Vartype.SPIN

    Note:
        The decorated function can explicitly list (name) vartype arguments
        constrained by :func:`vartype_argument` or it can use a keyword
        arguments `dict`.

    See also:
        :func:`~dimod.as_vartype`

    """
    # by default, constrain only one argument, the 'vartype`
    if not arg_names:
        arg_names = ['vartype']

    def _vartype_arg(f):
        argspec = inspect.getfullargspec(f)

        def _enforce_single_arg(name, args, kwargs):
            try:
                vartype = kwargs[name]
            except KeyError:
                raise TypeError('vartype argument missing')

            kwargs[name] = as_vartype(vartype)

        @wraps(f)
        def new_f(*args, **kwargs):
            # bound actual f arguments (including defaults) to f argument names
            # (note: if call arguments don't match actual function signature,
            # we'll fail here with the standard `TypeError`)
            bound_args = inspect.getcallargs(f, *args, **kwargs)

            # `getcallargs` doesn't merge additional positional/keyword arguments,
            # so do it manually
            final_args = list(bound_args.pop(argspec.varargs, ()))
            final_kwargs = bound_args.pop(argspec.varkw, {})

            final_kwargs.update(bound_args)
            for name in arg_names:
                _enforce_single_arg(name, final_args, final_kwargs)

            return f(*final_args, **final_kwargs)

        return new_f

    return _vartype_arg


def _is_integer(a):
    if isinstance(a, int):
        return True
    if hasattr(a, "is_integer") and a.is_integer():
        return True
    return False


# we would like to do graph_argument(*arg_names, allow_None=False), but python2...
def graph_argument(*arg_names, **options):
    """Decorator to coerce given graph arguments into a consistent form.

    The wrapped function accepts either an integer n, interpreted as a
    complete graph of size n, a nodes/edges pair, a sequence of edges, or a
    NetworkX graph. The argument is converted into a nodes/edges 2-tuple.

    Args:
        *arg_names (optional, default='G'):
            Names of the arguments for input graphs.

        allow_None (bool, optional, default=False):
            If True, None can be passed through as an input graph.

    """

    # by default, constrain only one argument, the 'G`
    if not arg_names:
        arg_names = ['G']

    # we only allow one option allow_None
    allow_None = options.pop("allow_None", False)
    if options:
        # to keep it consistent with python3
        # behaviour like graph_argument(*arg_names, allow_None=False)
        key, _ = options.popitem()
        msg = "graph_argument() for an unexpected keyword argument '{}'".format(key)
        raise TypeError(msg)

    def _graph_arg(f):
        argspec = inspect.getfullargspec(f)

        def _enforce_single_arg(name, args, kwargs):
            try:
                G = kwargs[name]
            except KeyError:
                raise TypeError('Graph argument missing')

            if hasattr(G, 'edges') and hasattr(G, 'nodes'):
                # networkx or perhaps a named tuple
                kwargs[name] = (list(G.nodes), list(G.edges))

            elif _is_integer(G):
                # an integer, cast to a complete graph
                kwargs[name] = (list(range(G)), list(itertools.combinations(range(G), 2)))

            elif isinstance(G, abc.Sequence):
                if len(G) != 2:
                    # edgelist
                    kwargs[name] = (list(set().union(*G)), G)
                else:  # len(G) == 2
                    # need to determine if this is a nodes/edges pair or an
                    # edgelist
                    if isinstance(G[0], int):
                        # nodes are an int so definitely nodelist
                        kwargs[name] = (list(range(G[0])), G[1])
                    elif all(isinstance(e, abc.Sequence) and len(e) == 2
                             for e in G):
                        # ok, everything is a sequence and everything has length
                        # 2, so probably an edgelist. But we're dealing with
                        # only four objects so might as well check to be sure
                        nodes, edges = G
                        if all(isinstance(e, abc.Sequence) and len(e) == 2 and
                               (v in nodes for v in e) for e in edges):
                            pass  # nodes, edges
                        else:
                            # edgelist
                            kwargs[name] = (list(set().union(*G)), G)
                    else:
                        # nodes, edges
                        pass

            elif allow_None and G is None:
                # allow None to be passed through
                kwargs[name] = G

            else:
                raise ValueError('Unexpected graph input form')

            return

        @wraps(f)
        def new_f(*args, **kwargs):
            # bound actual f arguments (including defaults) to f argument names
            # (note: if call arguments don't match actual function signature,
            # we'll fail here with the standard `TypeError`)
            bound_args = inspect.getcallargs(f, *args, **kwargs)

            # `getcallargs` doesn't merge additional positional/keyword arguments,
            # so do it manually
            final_args = list(bound_args.pop(argspec.varargs, ()))
            final_kwargs = bound_args.pop(argspec.varkw, {})

            final_kwargs.update(bound_args)
            for name in arg_names:
                _enforce_single_arg(name, final_args, final_kwargs)

            return f(*final_args, **final_kwargs)

        return new_f

    return _graph_arg


_NOT_FOUND = object()


def forwarding_method(func):
    """Improve the performance of a forwarding method by avoiding an attribute
    lookup.

    The decorated method should return the function that it is forwarding to.
    Subsequent calls will be made directly to that function.

    Example:

        >>> import typing
        >>> import timeit
        >>> from dimod.decorators import forwarding_method
        ...
        >>> class Inner:
        ...     def func(self, a: int, b: int = 0) -> int:
        ...         "Inner.func docsting."
        ...         return a + b
        ...
        >>> class Outer:
        ...     def __init__(self):
        ...         self.inner = Inner()
        ...
        ...     def func(self, a: int, b: int = 0) -> int:
        ...         "Outer.func docsting."
        ...         return self.inner.func(a, b=b)
        ...
        ...     @forwarding_method
        ...     def fwd_func(self, a: int, b: int = 0) -> int:
        ...         "Outer.fwd_func docsting."
        ...         return self.inner.func
        ...
        >>> obj = Outer()
        >>> obj.func(2, 3)
        5
        >>> obj.fwd_func(1, 3)
        4
        >>> timeit.timeit(lambda: obj.func(10, 5))  # doctest:+SKIP
        0.275462614998105
        >>> timeit.timeit(lambda: obj.fwd_func(10, 5))  # doctest:+SKIP
        0.16692455199881806
        >>> Outer.fwd_func.__doc__
        'Outer.fwd_func docsting.'
        >>> obj.fwd_func.__doc__
        'Inner.func docsting.'

    """
    @wraps(func)
    def wrapper(obj, *args, **kwargs):
        name = func.__name__

        try:
            cache = obj.__dict__
        except AttributeError:
            raise TypeError(
                f"No '__dict__' attribute on {type(obj).__name__!r}") from None

        method = cache.get(name, _NOT_FOUND)

        if method is _NOT_FOUND:
            # the args and kwargs are ignored but they are required to not
            # raise an error
            method = func(obj, *args, **kwargs)
            try:
                cache[name] = method
            except TypeError:
                raise TypeError(
                    f"the '__dict__' attribute of {type(obj).__name__!r} "
                    "instance does not support item assignment.") from None

        return method(*args, **kwargs)

    return wrapper

def unique_variable_labels(f):
    """Decorator to assign unique labels to variables when no label is passed.

    Designed to be applied to variable methods, :meth:`.dimod.Binary`,
    :meth:`.dimod.Spin`, :meth:`.dimod.Integer`.

    """
    @wraps(f)
    def conditional_unique_label(label = None, *args, **kwargs):
        if label is None:
            qm = f(label=new_label(), *args, **kwargs)
            return qm
        qm = f(label, *args, **kwargs)
        return qm

    return conditional_unique_label



# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Un_lt required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import typing

cimport cython

import numpy as np
cimport numpy as np

from dimod.libcpp cimport cppvartype_info
from dimod.typing import VartypeLike
from dimod.vartypes import as_vartype, Vartype

__all__ = ['vartype_info']


cdef extern from "numpy/arrayobject.h":
    ctypedef struct PyArray_Descr:
        pass

    object PyArray_Scalar(void*, PyArray_Descr*, object)

np.import_array()  # needed for PyArray_Scalar

# preconstruct these dtypes for speed, we could possibly improve it more
# by casting to PyArray_Descr but I had trouble getting that to work
cdef object _float32_dtype = np.dtype(np.float32)
cdef object _float64_dtype = np.dtype(np.float64)

cdef object as_numpy_float(cython.floating a):
    if cython.floating == double:
        return PyArray_Scalar(&a, <PyArray_Descr*>_float64_dtype, None)
    elif cython.floating == float:
        return PyArray_Scalar(&a, <PyArray_Descr*>_float32_dtype, None)
    else:
        raise NotImplementedError

@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline bint _lt(Integer[:] row, Integer[:] col, Py_ssize_t i, Py_ssize_t j):
    """Return (row[i], col[i]) < (row[j], col[j])."""
    return row[i] < row[j] or (row[i] == row[j] and col[i] < col[j])

@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline void _swap(Integer[:] row, Integer[:] col, cython.floating[:] data,
                       Py_ssize_t i, Py_ssize_t j):
    """Swap the values in the arrays at i, j"""
    row[i], row[j] = row[j], row[i]
    col[i], col[j] = col[j], col[i]
    data[i], data[j] = data[j], data[i]

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef Py_ssize_t coo_sort(Integer[:] row, Integer[:] col, cython.floating[:] data) except -1:
    """Sort COO arrays in-place by row then by column."""
    cdef Py_ssize_t num_interactions = row.shape[0]
    if col.shape[0] != num_interactions or data.shape[0] != num_interactions:
        raise ValueError("vectors should all be the same length")

    # row index should be _lt than col index, this handles upper-triangular vs
    # lower-triangular
    cdef Py_ssize_t i
    for i in range(num_interactions):
        if row[i] > col[i]:
            row[i], col[i] = col[i], row[i]

    quicksort_coo(row, col, data, 0, num_interactions - 1)


cdef void quicksort_coo(Integer[:] row, Integer[:] col, cython.floating[:] data,
                        Py_ssize_t low, Py_ssize_t high):
    # nb: high is inclusive
    cdef Py_ssize_t p
    if low < high:
        p = partition_coo(row, col, data, low, high)
        quicksort_coo(row, col, data, low, p - 1)
        quicksort_coo(row, col, data, p + 1, high)


cdef Py_ssize_t partition_coo(Integer[:] row, Integer[:] col, cython.floating[:] data,
                              Py_ssize_t low, Py_ssize_t high):

    # median of three pivot
    cdef Py_ssize_t mid = low + (high - low) // 2
    if _lt(row, col, mid, low):
        _swap(row, col, data, low, mid)
    if _lt(row, col, high, low):
        _swap(row, col, data, low, high)
    if _lt(row, col, mid, high):
        _swap(row, col, data, mid, high)

    cdef Py_ssize_t pi = high    

    cdef Py_ssize_t i = low

    cdef Py_ssize_t j
    for j in range(low, high):
        if _lt(row, col, j, pi):
            # ok, smaller than pivot
            _swap(row, col, data, i, j)

            i += 1

    _swap(row, col, data, i, pi)

    return i


cdef cppVartype cppvartype(vartype) except? cppVartype.SPIN:
    if vartype is Vartype.SPIN:
        return cppVartype.SPIN
    elif vartype is Vartype.BINARY:
        return cppVartype.BINARY
    elif vartype is Vartype.INTEGER:
        return cppVartype.INTEGER
    elif vartype is Vartype.REAL:
        return cppVartype.REAL
    else:
        raise TypeError(f"unexpected vartype {vartype!r}")


# todo: type annotations, fix docs. This needs a followup PR
def vartype_info(vartype, dtype=np.float64):
    """Information about the variable bounds by variable type.

    Args:
        vartype:
            Variable type. One of:

                * :class:`~dimod.Vartype.SPIN`, ``'SPIN'``, ``{-1, 1}``
                * :class:`~dimod.Vartype.BINARY`, ``'BINARY'``, ``{0, 1}``
                * :class:`~dimod.Vartype.INTEGER`, ``'INTEGER'``

        dtype: One of :class:`~numpy.float64` and :class:`~numpy.float32`.

    Returns:
        A named tuple with ``default_min``, ``default_max``, ``min``, and
        ``max`` fields. These specify the default and largest bounds for
        the given variable type.

    """
    cdef cppVartype vt = cppvartype(as_vartype(vartype, extended=True))

    Info = typing.NamedTuple("VartypeLimits",
                      (("default_min", float),
                       ("default_max", float),
                       ("min", float),
                       ("max", float),
                      ))

    dtype = np.dtype(dtype)
    if dtype == np.dtype(np.float32):
        return Info(
            as_numpy_float(cppvartype_info[np.float32_t].default_min(vt)),
            as_numpy_float(cppvartype_info[np.float32_t].default_max(vt)),
            as_numpy_float(cppvartype_info[np.float32_t].min(vt)),
            as_numpy_float(cppvartype_info[np.float32_t].max(vt)),
            )
    elif dtype == np.dtype(np.float64):
        return Info(
            as_numpy_float(cppvartype_info[np.float64_t].default_min(vt)),
            as_numpy_float(cppvartype_info[np.float64_t].default_max(vt)),
            as_numpy_float(cppvartype_info[np.float64_t].min(vt)),
            as_numpy_float(cppvartype_info[np.float64_t].max(vt)),
            )
    else:
        raise ValueError("only supports np.float64 and np.float32 dtypes")



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
The :class:`.Sampler` abstract base class (:mod:`abc`) helps you create new
dimod samplers.

Any new dimod sampler must define a subclass of :class:`.Sampler` that implements
abstract properties :attr:`~.Sampler.parameters` and :attr:`~.Sampler.properties`
and one of the abstract methods :meth:`~.Sampler.sample`, :meth:`~.Sampler.sample_ising`,
or :meth:`~.Sampler.sample_qubo`. The :class:`.Sampler` class provides the complementary
methods as mixins and ensures consistent responses.

Implemented sample methods must accept, and warn on, unknown keyword arguments
`**kwargs`. This means that all implemented sample methods must have the
`**kwargs` parameter. :meth:`~.Sampler.remove_unknown_kwargs` is a convenience
method provided for this purpose.

For example, the following steps show how to easily create a dimod sampler. It is
sufficient to implement a single method (in this example the
:meth:`~.Sampler.sample_ising` method) to create a dimod sampler with the
:class:`.Sampler` class.

.. testcode::

    class LinearIsingSampler(dimod.Sampler):

        def sample_ising(self, h, J, **kwargs):
            kwargs = self.remove_unknown_kwargs(**kwargs)
            sample = linear_ising(h, J)
            energy = dimod.ising_energy(sample, h, J)
            return dimod.SampleSet.from_samples([sample], vartype='SPIN', energy=[energy])

        @property
        def properties(self):
            return dict()

        @property
        def parameters(self):
            return dict()

For this example, the implemented sampler :meth:`~.Sampler.sample_ising` can be based on
a simple placeholder function, which returns a sample that minimizes the linear terms:

.. testcode::

    def linear_ising(h, J):
        sample = {}
        for v in h:
            if h[v] < 0:
                sample[v] = +1
            else:
                sample[v] = -1
        return sample


The :class:`.Sampler` ABC provides the other sample methods "for free"
as mixins.

>>> sampler = LinearIsingSampler()
...
... # Implemented by class LinearIsingSampler:
>>> response = sampler.sample_ising({'a': -1}, {})
...
...  # Mixins provided by Sampler class:
>>> sampleset = sampler.sample_qubo({('a', 'a'): 1})
>>> sampleset = sampler.sample(dimod.BinaryQuadraticModel.from_ising({'a': -1}, {}))

Below is a more complex version of the same sampler, where the
:attr:`~.Sampler.properties` and :attr:`~.Sampler.parameters` properties return
non-empty dicts.

.. testcode::

    class FancyLinearIsingSampler(dimod.Sampler):
        def __init__(self):
            self._properties = {'description': 'a simple sampler that only considers the linear terms'}
            self._parameters = {'verbose': []}

        def sample_ising(self, h, J, verbose=False, **kwargs):
            kwargs = self.remove_unknown_kwargs(**kwargs)
            sample = linear_ising(h, J)
            energy = dimod.ising_energy(sample, h, J)
            if verbose:
                print(sample)
            return dimod.SampleSet.from_samples([sample], energy=[energy])

        @property
        def properties(self):
            return self._properties

        @property
        def parameters(self):
            return self._parameters

"""

import abc
import typing
import warnings

from dimod.binary_quadratic_model import BinaryQuadraticModel
from dimod.exceptions import InvalidSampler, SamplerUnknownArgWarning
from dimod.sampleset import SampleSet
from dimod.typing import Bias, Variable
from dimod.vartypes import Vartype

__all__ = ['Sampler']


class SamplerABCMeta(abc.ABCMeta):
    def __new__(mcls, name, bases, namespace, **kwargs):
        cls = abc.ABCMeta.__new__(mcls, name, bases, namespace, **kwargs)

        samplermixins = {name
                         for name, value in namespace.items()
                         if getattr(value, "__issamplemixin__", False)}
        if len(samplermixins) == 3:
            abstracts = samplermixins
        else:
            abstracts = set()

        for base in bases:
            samplermixins = {name
                             for name in getattr(base, "__abstractmethods__", set())
                             if getattr(getattr(cls, name, None), "__issamplemixin__", False)}
            if len(samplermixins) == 3:
                abstracts.update(samplermixins)

        # if we found any, update abstract methods
        if abstracts:
            cls.__abstractmethods__ = frozenset(abstracts.union(cls.__abstractmethods__))

        return cls


def samplemixinmethod(method):
    """Marks a method as being a mixin.

    Adds the '__issamplemixin__' attribute with value True to the decorated function.

    Examples:
        >>> @samplemixinmethod
        >>> def f():
        ...     pass
        >>> f.__issamplemixin__
        True

    """
    # NB: decorator name was chosen to be consistent with @classmethod and @staticmethod
    method.__issamplemixin__ = True
    return method


class Sampler(metaclass=SamplerABCMeta):
    """Abstract base class for dimod samplers.

    Provides all methods :meth:`~.Sampler.sample`, :meth:`~.Sampler.sample_ising`,
    :meth:`~.Sampler.sample_qubo` assuming at least one is implemented.

    Also includes utility method :meth:`~.Sampler.remove_unknown_kwargs`,
    which may be used in sample methods to handle unknown kwargs.
    """

    @abc.abstractproperty  # for python2 compatibility
    def parameters(self) -> typing.Dict[str, typing.Any]:
        """Parameters as a dict, where keys are keyword parameters accepted by the
        sampler methods and values are lists of the properties relevent to each
        parameter.
        """
        pass

    @abc.abstractproperty  # for python2 compatibility
    def properties(self) -> typing.Dict[str, typing.Any]:
        """Properties as a dict containing any additional information about the
        sampler.
        """
        pass

    @samplemixinmethod
    def sample(self, bqm: BinaryQuadraticModel, **parameters) -> SampleSet:
        """Sample from a binary quadratic model.

        This method is inherited from the :class:`.Sampler` base class.

        Converts the binary quadratic model to either Ising or QUBO format and
        then invokes an implemented sampling method (one of
        :meth:`.sample_ising` or :meth:`.sample_qubo`).

        Args:

            bqm: A binary quadratic model.

            **parameters:
                See the implemented sampling for additional keyword definitions.
                Unknown keywords are accepted with a warning raised.

        Returns: Samples from the binary quadratic model.

        See also:
            :meth:`.sample_ising`, :meth:`.sample_qubo`

        """

        # we try to use the matching sample method if possible
        if bqm.vartype is Vartype.SPIN:
            if not getattr(self.sample_ising, '__issamplemixin__', False):
                # sample_ising is implemented
                h, J, offset = bqm.to_ising()
                sampleset = self.sample_ising(h, J, **parameters)
            else:
                Q, offset = bqm.to_qubo()
                sampleset = self.sample_qubo(Q, **parameters)
        elif bqm.vartype is Vartype.BINARY:
            if not getattr(self.sample_qubo, '__issamplemixin__', False):
                # sample_qubo is implemented
                Q, offset = bqm.to_qubo()
                sampleset = self.sample_qubo(Q, **parameters)
            else:
                h, J, offset = bqm.to_ising()
                sampleset = self.sample_ising(h, J, **parameters)
        else:
            raise RuntimeError("binary quadratic model has an unknown vartype")

        # if the vartype already matches this will just adjust the offset
        return sampleset.change_vartype(bqm.vartype, energy_offset=offset)

    @samplemixinmethod
    def sample_ising(self, h: typing.Union[typing.Mapping[Variable, Bias], typing.Sequence[Bias]],
                     J: typing.Mapping[typing.Tuple[Variable, Variable], Bias],
                     **parameters) -> SampleSet:
        """Sample from an Ising model using the implemented sample method.

        This method is inherited from the :class:`.Sampler` base class.

        Converts the Ising model into a :obj:`.BinaryQuadraticModel` and then
        calls :meth:`.sample`.

        Args:
            h: Linear biases of the Ising problem. If a list, indices are the
                variable labels.

            J: Quadratic biases of the Ising problem.

            **kwargs:
                See the implemented sampling for additional keyword definitions.

        Returns: Samples from the Ising model.

        See also:
            :meth:`.sample`, :meth:`.sample_qubo`

        """
        bqm = BinaryQuadraticModel.from_ising(h, J)
        return self.sample(bqm, **parameters)

    @samplemixinmethod
    def sample_qubo(self, Q: typing.Mapping[typing.Tuple[Variable, Variable], Bias],
                    **parameters)  -> SampleSet:
        """Sample from a QUBO using the implemented sample method.

        This method is inherited from the :class:`.Sampler` base class.

        Converts the quadratic unconstrained binary optimization (QUBO) into a
        :obj:`.BinaryQuadraticModel` and then calls :meth:`.sample`.

        Args:
            Q: Coefficients of a QUBO problem.

            **kwargs:
                See the implemented sampling for additional keyword definitions.

        Returns: Samples from a QUBO.

        See also:
            :meth:`.sample`, :meth:`.sample_ising`

        """
        bqm = BinaryQuadraticModel.from_qubo(Q)
        return self.sample(bqm, **parameters)

    def remove_unknown_kwargs(self, **kwargs) -> typing.Dict[str, typing.Any]:
        """Remove with warnings any keyword arguments not accepted by the sampler.

        Args:
            **kwargs:
                Keyword arguments to be validated.

        Returns: Updated `kwargs` dict.

        Examples:
            >>> import warnings
            >>> sampler = dimod.RandomSampler()
            >>> with warnings.catch_warnings():
            ...     warnings.filterwarnings('ignore')
            ...     try:
            ...         kwargs = sampler.remove_unknown_kwargs(num_reads=10,
            ...                                                non_param=3)
            ...     except dimod.exceptions.SamplerUnknownArgWarning:
            ...:        pass
            {'num_reads': 10}
        """
        for kw in [k for k in kwargs if k not in self.parameters]:
            msg = "Ignoring unknown kwarg: {!r}".format(kw)
            warnings.warn(msg, SamplerUnknownArgWarning, stacklevel=3)
            kwargs.pop(kw)

        return kwargs



# Copyright 2022 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os

from cython.operator cimport preincrement as inc, dereference as deref
from libcpp.string cimport string
from libcpp.unordered_map cimport unordered_map
from libcpp.vector cimport vector

import numpy as np
cimport numpy as np

from dimod.constrained import ConstrainedQuadraticModel
from dimod.libcpp cimport cppread_lp, cppLPModel, cppQuadraticModel
from dimod.quadratic cimport cyQM_float64
from dimod.quadratic.cyqm.cyqm_float64 cimport bias_type, index_type
from dimod.quadratic.cyqm.cyqm_float64 import BIAS_DTYPE
from dimod.quadratic import QuadraticModel


cdef void _swap_qm(cyQM_float64 qm, cppQuadraticModel[bias_type, index_type]& cppqm, unordered_map[string, index_type] labels) except +:
    assert cppqm.num_variables() == labels.size()
    assert qm.num_variables() == 0

    qm.cppqm.swap(cppqm)

    cdef vector[string] variables
    variables.resize(labels.size())
    it = labels.begin()
    while it != labels.end():
        variables[deref(it).second] = deref(it).first
        inc(it)

    # cdef object v
    vit = variables.begin()
    while vit != variables.end():
        qm.variables._append(deref(vit).decode())
        inc(vit)


def cyread_lp_file(object filename):
    """Create a constrained quadratic model from the given LP file."""

    if not os.path.isfile(filename):
        raise ValueError(f"no file named {filename}")

    cdef string _filename = filename.encode()
    cdef cppLPModel[bias_type, index_type] lpmodel = cppread_lp[bias_type, index_type](_filename)

    cqm = ConstrainedQuadraticModel()

    if cqm.objective.dtype != BIAS_DTYPE:
        raise RuntimeError("unexpected dtype")

    _swap_qm(cqm.objective.data, lpmodel.objective.model, lpmodel.objective.labels)

    cdef Py_ssize_t i
    for i in range(lpmodel.constraints.size()):
        lhs = QuadraticModel(dtype=BIAS_DTYPE)
        _swap_qm(lhs.data, lpmodel.constraints[i].lhs.model, lpmodel.constraints[i].lhs.labels)

        cqm.add_constraint_from_model(
            lhs,
            sense=lpmodel.constraints[i].sense.decode(),
            rhs=lpmodel.constraints[i].rhs,
            label=lpmodel.constraints[i].lhs.name.decode(),
            copy=False,
            )

    return cqm



# Copyright 2022 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from __future__ import annotations

import collections.abc
import itertools
import typing

import numpy as np

import dimod  # for typing

from dimod.binary import BinaryQuadraticModel
from dimod.vartypes import Vartype


__all__ = ["random_nae3sat", "random_2in4sat"]


def _kmcsat_interactions(num_variables: int, k: int, num_clauses: int,
                         *,
                         seed: typing.Union[None, int, np.random.Generator] = None,
                         ) -> typing.Iterator[typing.Tuple[int, int, int]]:
    rng = np.random.default_rng(seed)

    for _ in range(num_clauses):
        # randomly select the variables
        variables = rng.choice(num_variables, k, replace=False)

        # randomly assign the negations
        signs = 2 * rng.integers(0, 1, endpoint=True, size=k) - 1

        # get the interactions for each clause
        for (u, usign), (v, vsign) in itertools.combinations(zip(variables, signs), 2):
            yield u, v, usign*vsign


def random_kmcsat(variables: typing.Union[int, typing.Sequence[dimod.typing.Variable]],
                  k: int,
                  num_clauses: int,
                  *,
                  seed: typing.Union[None, int, np.random.Generator] = None,
                  ) -> BinaryQuadraticModel:
    """Generate a random k Max-Cut satisfiability problem as a binary quadratic model.

    kMC-SAT [#]_ is an NP-complete problem class
    that consists in satisfying a number of conjunctive
    clauses of ``k`` literals (variables, or their negations).
    Each clause should encode a max-cut problem over the clause literals.

    .. [#] Adam Douglass, Andrew D. King & Jack Raymond,
       "Constructing SAT Filters with a Quantum Annealer",
       https://link.springer.com/chapter/10.1007/978-3-319-24318-4_9

    Args:
        num_variables: The number of variables in the problem.
        num_clauses: The number of clauses. Each clause contains three literals.
        seed: Passed to :func:`numpy.random.default_rng()`, which is used
            to generate the clauses and the variable negations.

    Returns:
        A binary quadratic model with spin variables.

    .. note:: The clauses are randomly sampled from the space of 4-variable
        clauses *with replacement* which can result in collisions. However,
        collisions are allowed in standard problem definitions, are absent with
        high probability in interesting cases, and are almost always harmless
        when they do occur.

    """
    if isinstance(variables, collections.abc.Sequence):
        num_variables = len(variables)
        labels = variables
    else:
        num_variables = variables
        labels = None

    if num_variables < k:
        raise ValueError(f"must use at least {k} variables")

    if num_clauses < 0:
        raise ValueError("num_clauses must be non-negative")

    bqm = BinaryQuadraticModel(num_variables, Vartype.SPIN)
    bqm.add_quadratic_from(_kmcsat_interactions(num_variables, k, num_clauses, seed=seed))

    if labels:
        bqm.relabel_variables(dict(enumerate(labels)))

    return bqm


def random_nae3sat(variables: typing.Union[int, typing.Sequence[dimod.typing.Variable]],
                   num_clauses: int,
                   *,
                   seed: typing.Union[None, int, np.random.Generator] = None,
                   ) -> BinaryQuadraticModel:
    """Generate a random not-all-equal 3-satisfiability problem as a binary quadratic model.

    Not-all-equal 3-satisfiability (NAE3SAT_) is an NP-complete problem class
    that consists in satisfying a number of conjunctive
    clauses of three literals (variables, or their negations).
    For valid solutions, the literals in each clause should be not-all-equal;
    i.e. any assignment of values except ``(+1, +1, +1)`` or ``(-1, -1, -1)``
    are valid for each clause.

    NAE3SAT problems have been studied with the D-Wave quantum annealer [#]_.

    .. _NAE3SAT: https://en.wikipedia.org/wiki/Not-all-equal_3-satisfiability

    .. [#] Adam Douglass, Andrew D. King & Jack Raymond,
       "Constructing SAT Filters with a Quantum Annealer",
       https://link.springer.com/chapter/10.1007/978-3-319-24318-4_9

    Args:
        num_variables: The number of variables in the problem.
        num_clauses: The number of clauses. Each clause contains three literals.
        seed: Passed to :func:`numpy.random.default_rng()`, which is used
            to generate the clauses and the variable negations.

    Returns:
        A binary quadratic model with spin variables.

    Example:

        Generate a NAE3SAT problem with a given clause-to-variable ratio (rho).

        >>> num_variables = 75
        >>> rho = 2.1
        >>> bqm = dimod.generators.random_nae3sat(num_variables, round(num_variables*rho))

    .. note:: The clauses are randomly sampled from the space of 3-variable
        clauses *with replacement* which can result in collisions. However,
        collisions are allowed in standard problem definitions, are absent with
        high probability in interesting cases, and are almost always harmless
        when they do occur.

    """
    return random_kmcsat(variables, 3, num_clauses, seed=seed)


def random_2in4sat(variables: typing.Union[int, typing.Sequence[dimod.typing.Variable]],
                   num_clauses: int,
                   *,
                   seed: typing.Union[None, int, np.random.Generator] = None,
                   ) -> BinaryQuadraticModel:
    """Generate a random 2-in-4 satisfiability problem as a binary quadratic model.

    2-in-4 satisfiability [#]_ is an NP-complete problem class
    that consists in satisfying a number of conjunctive
    clauses of four literals (variables, or their negations).
    For valid solutions, two of the literals in each clause should ``+1`` and
    the other two should be ``-1``.

    .. [#] Adam Douglass, Andrew D. King & Jack Raymond,
       "Constructing SAT Filters with a Quantum Annealer",
       https://link.springer.com/chapter/10.1007/978-3-319-24318-4_9

    Args:
        num_variables: The number of variables in the problem.
        num_clauses: The number of clauses. Each clause contains three literals.
        seed: Passed to :func:`numpy.random.default_rng()`, which is used
            to generate the clauses and the variable negations.

    Returns:
        A binary quadratic model with spin variables.

    .. note:: The clauses are randomly sampled from the space of 4-variable
        clauses *with replacement* which can result in collisions. However,
        collisions are allowed in standard problem definitions, are absent with
        high probability in interesting cases, and are almost always harmless
        when they do occur.

    """
    return random_kmcsat(variables, 4, num_clauses, seed=seed)
